# -------------------------------------------------------------------------
#
#  Part of the CodeChecker project, under the Apache License v2.0 with
#  LLVM Exceptions. See LICENSE for license information.
#  SPDX-License-Identifier: Apache-2.0 WITH LLVM-exception
#
# -------------------------------------------------------------------------

# The supported formats the users should specify. (This is not an exhaustive
# list of ALL formats available.)
USER_FORMATS = ['rows', 'table', 'csv', 'json']
